"""STREAMLINED OFF-POLICY LEARNING

From the paper:
TOWARDS SIMPLICITY IN DEEP REINFORCEMENT LEARNING: STREAMLINED OFF-POLICY LEARNING
https://arxiv.org/abs/1910.02208.
"""
from .policy import SOPTorchPolicy
from .trainer import SOPTrainer
